/* 
 * PPRacer 
 * Copyright (C) 2003-2004 The PPRacer Team
 *
 * Copyright (C) 1999-2001 Jasmin F. Patry
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "checkbox.h"
 
#include "textarea.h"
#include "ui_mgr.h"

#include "render_util.h"
#include "string_util.h"

#define DEFAULT_ARROW_BUTTON_HEIGHT 15
#define DEFAULT_ARROW_BUTTON_WIDTH  27
#define DEFAULT_ARROW_REGION_WIDTH  36
#define DEFAULT_ARROW_VERT_SEPARATION  4
#define DEFAULT_BORDER_WIDTH 4
#define DEFAULT_TEXT_PAD 4

namespace pp {

Textarea::Textarea( pp::Vec2d pos, pp::Vec2d size,
			     char *font_binding, char *text )
{
    char *binding;
    pp::Vec2d ll, ur;
	
	this->position = pos;
	this->size = size;
	
    this->arrow_width = DEFAULT_ARROW_REGION_WIDTH;
    this->border_width = DEFAULT_BORDER_WIDTH;
    this->text_pad = DEFAULT_TEXT_PAD;
	
    calcTextRegionDims();

    this->font_binding = font_binding;
    this->border_colour = theme.foreground;
    this->background_colour = theme.background;

    /* 
     * Create buttons 
     */
    this->up_button = new pp::Button(
		pp::Vec2d( 0, 0 ), /* position will be set later */
		pp::Vec2d(DEFAULT_ARROW_BUTTON_WIDTH, DEFAULT_ARROW_BUTTON_HEIGHT),
		NULL,
		NULL );

    binding = "textarea_arrows";

    ll = pp::Vec2d( 2.0/64.0, 17.0/64.0 );
    ur = pp::Vec2d( 29.0/64.0, 32.0/64.0 );
    this->up_button->setImage( binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 34.0/64.0, 17.0/64.0 );
    ur = pp::Vec2d( 61.0/64.0, 32.0/64.0 );
    this->up_button->setDisabledImage( binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 34.0/64.0, 49.0/64.0 );
    ur = pp::Vec2d( 61.0/64.0, 64.0/64.0 );
    this->up_button->setHilitImage( binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 2.0/64.0, 49.0/64.0 );
    ur = pp::Vec2d( 29.0/64.0, 64.0/64.0 );
    this->up_button->setClickedImage( binding, ll, ur, pp::Color::white );
	
	this->up_button->setActive(true);
    this->up_button->signalClicked.Connect(pp::CreateSlot(this,&Textarea::upButtonClick));

    this->down_button = new pp::Button(
		pp::Vec2d( 0, 0 ), /* position will be set later */
		pp::Vec2d(DEFAULT_ARROW_BUTTON_WIDTH,DEFAULT_ARROW_BUTTON_HEIGHT),
		NULL,
		NULL );

    binding = "textarea_arrows";

    ll = pp::Vec2d( 2.0/64.0, 1.0/64.0 );
    ur = pp::Vec2d( 29.0/64.0, 16.0/64.0 );
    this->down_button->setImage(binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 34.0/64.0, 1.0/64.0 );
    ur = pp::Vec2d( 61.0/64.0, 16.0/64.0 );
    this->down_button->setDisabledImage(binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 34.0/64.0, 33.0/64.0 );
    ur = pp::Vec2d( 61.0/64.0, 48.0/64.0 );
    this->down_button->setHilitImage(binding, ll, ur, pp::Color::white );

    ll = pp::Vec2d( 2.0/64.0, 33.0/64.0 );
    ur = pp::Vec2d( 29.0/64.0, 48.0/64.0 );
    this->down_button->setClickedImage(binding, ll, ur, pp::Color::white );

 	this->down_button->setActive(true);
    this->down_button->signalClicked.Connect(pp::CreateSlot(this,&Textarea::downButtonClick));

	updateButtonPositions();
    

    if ( text == NULL ) {
		text = "";
    }
    this->text_orig = string_copy( text );
    //this->line_list = NULL;
    this->text_lines = NULL;

    createLines();

    updateButtonEnabledStates();

    this->visible = false;
    this->active = false;
}

Textarea::~Textarea()
{
    setVisible( false );
    setActive( false );

    if ( up_button != NULL ) {
		delete up_button;
    }
    if ( down_button != NULL ) {
		delete down_button;
    }

    if ( text_lines != NULL ) {
		free( text_lines );
    }

    if ( text_orig != NULL ) {
		free( text_orig );
    }	
}

void
Textarea::calcTextRegionDims()
{
    text_region_width = size.x - 2*border_width - 2*text_pad - arrow_width;
    text_region_height = size.y - 2*border_width - 2*text_pad;
}

void
Textarea::updateButtonPositions()
{
	up_button->setPosition(pp::Vec2d(
	    position.x + size.x - border_width -
	    up_button->getWidth(),
	    position.y + size.y - border_width -
	    up_button->getHeight() ) );
	
	down_button->setPosition(pp::Vec2d(
	    position.x + size.x - border_width -
	    down_button->getWidth(),
	    position.y + border_width ) );
}

void
Textarea::createLines()
{
    char *cur_ptr;
    char *old_space_ptr;
    char *space_ptr;

	
    font_t font;


    if ( !get_font_binding( font_binding, font ) ) {
		print_warning( IMPORTANT_WARNING,
		       "Couldn't get font for binding %s", font_binding );
		return;
    }

    if ( text_lines != NULL ) {
		free( text_lines );
		text_lines = NULL;
    }

    if ( !lines.empty() ) {
		lines.clear();
    }

    text_lines = string_copy( text_orig );

    cur_ptr = text_lines;

    while ( *cur_ptr != '\0' ){
		old_space_ptr = NULL;
		space_ptr = strchr( cur_ptr, ' ' );
		
		while (true) {

		    if ( space_ptr != NULL ) {
				*space_ptr = '\0';
		    }
		
	   		if ( string_width( cur_ptr, &font ) > text_region_width ) {
				break;
	    	}
		
	    	old_space_ptr = space_ptr;

	   		if ( space_ptr == NULL ) {
				/* Entire string fits in area */
				break;
		    }
	
		    *space_ptr = ' ';
	
		    space_ptr = strchr( space_ptr+1, ' ' );
		}

		if ( old_space_ptr == NULL ) {
			    /* Either string is too wide for area, or the entire 
		       remaining portion of string fits in area (space_ptr == NULL).
		    */
		    //cur_line = insert_list_elem( line_list, cur_line,
			//			 (list_elem_data_t) cur_ptr );
			lines.push_back(cur_ptr);
			
		    cur_ptr += strlen( cur_ptr );

		    if ( space_ptr != NULL ) {
				/* Advance past the NULL since there's more string left */
				cur_ptr += 1;
		    }
		} else {
		    if ( space_ptr != NULL ) {
			*space_ptr = ' ';
		    }
		    *old_space_ptr = '\0';
		    //cur_line = insert_list_elem( line_list, cur_line,
			//			 (list_elem_data_t) cur_ptr );
			lines.push_back(cur_ptr);
		    cur_ptr = old_space_ptr + 1; 
		}
    }

    top_line = lines.begin(); //get_list_head( line_list );
}

void
Textarea::updateButtonEnabledStates()
{
    if ( lines.empty() ) {
		/* No lines */
		up_button->setEnabled( false );
		down_button->setEnabled( false );
    } else {
		if ( top_line == lines.begin() ) {
		    up_button->setEnabled( false );
		} else {
		    up_button->setEnabled( true );
		}
		if ( top_line == --lines.end() ) {
		    down_button->setEnabled( false );
		} else {
		    down_button->setEnabled( true );
		}
    }
}

int
Textarea::string_width( char *string, font_t *font )
{
    int w, asc, desc;
    get_font_metrics( font, string, &w, &asc, &desc );
    return w;
}

void
Textarea::setText( char *text )
{
    if ( text_orig != NULL ) {
		free( text_orig );
		text_orig = NULL;
    }

    if ( text == NULL ) {
		text = "";
    }

    text_orig = string_copy( text );
    createLines();
    updateButtonEnabledStates();	
}

void
Textarea::setPosition(pp::Vec2d pos)
{
	Widget::setPosition(pos);
	updateButtonPositions();
}


void 
Textarea::upButtonClick()
{
	if( top_line != lines.begin()){
		top_line--;
    	updateButtonEnabledStates();
    	UIMgr.setDirty();
	}
}

void 
Textarea::downButtonClick()
{
	if( top_line != --lines.end() ){
		top_line++;
		updateButtonEnabledStates();
		UIMgr.setDirty();
	}
}

void
Textarea::drawTextLines( font_t *font )
{
    std::list<char*>::iterator cur_line;
    int cur_y;
    int w, asc, desc;
    char *line;
    int height_used;

    if ( lines.empty() ) {
		return;
    }
	
	cur_line = top_line;

    line = *(cur_line);

	if(line==0){
		return;
	}
	
	bind_font_texture( font );

    check_assertion( line != NULL, "line is NULL in line_list" );

    get_font_metrics( font, line, &w, &asc, &desc );

    glPushMatrix();
    {
	cur_y = int(position.y + size.y - border_width - text_pad - asc);
	glTranslatef( position.x + border_width + text_pad, cur_y, 0 );
	height_used = asc + desc;
	    
	while (true){
		
		if(line==0) break;
		
	    if ( height_used > text_region_height ) {
			if ( down_button ) {
		    	down_button->setEnabled( true );
			}
			break;
	    }

	    get_font_metrics( font, line, &w, &asc, &desc );
	    draw_string( font, line );

	    //cur_line++; // = get_next_list_elem( line_list, cur_line );

	    if ( cur_line == --lines.end()) {
			if ( down_button ) {
		    	down_button->setEnabled( false );
			}
			break;
	    }
		
		cur_line++; // = get_next_list_elem( line_list, cur_line );
	    line = *(cur_line);
		
	    glTranslatef( -w, -(asc+desc), 0 );

	    height_used += asc + desc;
	}
    }
    glPopMatrix();
}

void
Textarea::draw()
{
    font_t font;

    glDisable( GL_TEXTURE_2D );

    glColor3dv( (double*)&border_colour );

    glRectf( position.x, 
	     position.y,
	     position.x + size.x,
	     position.y + size.y );

    glColor3dv( (double*)&background_colour );

    glRectf( position.x + border_width, 
	     position.y + border_width,
	     position.x + size.x - border_width,
	     position.y + size.y - border_width );

    glEnable( GL_TEXTURE_2D );

    if ( !get_font_binding( font_binding, font ) ) {
		print_warning( IMPORTANT_WARNING,
		       "Couldn't get font object for binding %s",
		       font_binding );
    } else {
		drawTextLines( &font );
    }
    if ( up_button != NULL ) {
		up_button->draw();
    }

    if ( down_button != NULL ) {
		down_button->draw(); 
    }
}	


} //namespace pp
